/* Copyright 2013, Martin Koller, kollix@aon.at

  This file is part of airsyncDownload.

  airsyncDownload is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  airsyncDownload is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with airsyncDownload.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <Session.hxx>

#include <wbxml.h>
#include <iostream>

#include <QUuid>
#include <QSslError>

//--------------------------------------------------------------------------------

Session::Session(const QString &server, const QString &domain,
                 const QString &username, const QString &password,
                 const QString &uniqueDeviceId)
  : serverUrl(server), policyKey(0), wroteSslWarning(false)
{
  nam = new QNetworkAccessManager(this);
  eventLoop = new QEventLoop(this);

  connect(nam, SIGNAL(sslErrors(QNetworkReply *, const QList<QSslError> &)),
          this,  SLOT(sslErrors(QNetworkReply *, const QList<QSslError> &)));

  serverUrl.setPath(QLatin1String("/Microsoft-Server-ActiveSync"));
  // set query parts which stay the same
  serverUrl.addQueryItem("User", (domain + '\\' + username).toUtf8().toPercentEncoding());
  serverUrl.addQueryItem("DeviceId", uniqueDeviceId);
  serverUrl.addQueryItem("DeviceType", "SP");

  loginBase64 = (domain + '\\' + username + ':' + password).toUtf8().toBase64();
}

//--------------------------------------------------------------------------------

QByteArray Session::sendCommand(const QByteArray &cmd, const QByteArray &data)
{
  QNetworkRequest req;

  QUrl url = serverUrl;
  url.addQueryItem(QLatin1String("Cmd"), QString::fromUtf8(cmd));
  req.setUrl(url);

  return post(req, data);
}

//--------------------------------------------------------------------------------

void Session::options()
{
  std::cerr << "------------------------- getting OPTIONS" << std::endl;

  QNetworkRequest req;

  req.setUrl(serverUrl);
  req.setRawHeader("Authorization", QByteArray("Basic ") + loginBase64);

  if ( debug )
  {
    qWarning("%s", qPrintable(req.url().toString()));
    QList<QByteArray> rawHeaders = req.rawHeaderList();
    for (int i = 0; i < rawHeaders.count(); i++)
      qWarning("%s: %s", qPrintable(rawHeaders[i]), qPrintable(req.rawHeader(rawHeaders[i])));
  }

  QNetworkReply *reply = nam->sendCustomRequest(req, "OPTIONS");

  connect(reply, SIGNAL(finished()), eventLoop, SLOT(quit()));
  connect(reply, SIGNAL(error(QNetworkReply::NetworkError)), eventLoop, SLOT(quit()));

  eventLoop->exec();

  if ( reply->error() != QNetworkReply::NoError )
  {
    std::cerr << "ERROR:" << qPrintable(reply->errorString()) << std::endl;
    std::cerr << qPrintable(req.url().toString()) << std::endl;
  }
  else
  {
    QList<QByteArray> rawHeaders = reply->rawHeaderList();
    for (int i = 0; i < rawHeaders.count(); i++)
      std::cerr << qPrintable(rawHeaders[i]) << ": " << qPrintable(reply->rawHeader(rawHeaders[i])) << std::endl;
  }

  delete reply;
}

//--------------------------------------------------------------------------------

QByteArray Session::post(QNetworkRequest req, const QByteArray &data)
{
  const QByteArray xmlHeader =
    "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
    "<!DOCTYPE ActiveSync PUBLIC \"-//MICROSOFT//DTD ActiveSync//EN\" \"http://www.microsoft.com/\">\n";

  if ( debug )
  {
    qWarning("------------------------- posting");
    qWarning("%s", qPrintable(req.url().toString()));
  }

  req.setRawHeader("Authorization", QByteArray("Basic ") + loginBase64);
  req.setRawHeader("MS-ASProtocolVersion", "12.1");

  if ( policyKey > 0 )
    req.setRawHeader("X-MS-PolicyKey", QByteArray::number(policyKey));

  req.setHeader(QNetworkRequest::ContentTypeHeader, "application/vnd.ms-sync.wbxml");
  QByteArray wbxml;
  if ( !dataToWbXml(xmlHeader + data, wbxml) )
    return "";

  if ( debug )
  {
    QList<QByteArray> rawHeaders = req.rawHeaderList();
    for (int i = 0; i < rawHeaders.count(); i++)
      qWarning("SENDING: %s: %s", qPrintable(rawHeaders[i]), qPrintable(req.rawHeader(rawHeaders[i])));

    qWarning(qPrintable(data));
    qWarning("decoded again from wbxml; this is what the server sees");
    QByteArray test;
    dataFromWbXml(wbxml, test);
    qWarning(qPrintable(test));
  }

  QNetworkReply *reply = nam->post(req, wbxml);

  connect(reply, SIGNAL(finished()), eventLoop, SLOT(quit()));
  connect(reply, SIGNAL(error(QNetworkReply::NetworkError)), eventLoop, SLOT(quit()));

  if ( debug ) qWarning("sent ...");
  eventLoop->exec();
  if ( debug ) qWarning("received ...");

  QByteArray result = reply->readAll();

  if ( debug )
  {
    QList<QByteArray> rawHeaders = reply->rawHeaderList();
    for (int i = 0; i < rawHeaders.count(); i++)
      qWarning("RECEIVED: %s: %s", qPrintable(rawHeaders[i]), qPrintable(reply->rawHeader(rawHeaders[i])));

    qWarning("HTTP result code: %d", reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt());
  }

  if ( reply->error() != QNetworkReply::NoError )
  {
    std::cerr << "ERROR:" << qPrintable(reply->errorString()) << std::endl;
    std::cerr << qPrintable(req.url().toString()) << std::endl;
  }
  else
  {
    if ( !result.isEmpty() )
      dataFromWbXml(result, result);

    if ( debug ) qWarning("data:%s", qPrintable(result));
  }

  delete reply;

  return result;
}

//--------------------------------------------------------------------------------

void Session::sslErrors(QNetworkReply *reply, const QList<QSslError> &errors)
{
  reply->ignoreSslErrors();  // ignore all errors

  // log errors
  if ( debug || !wroteSslWarning )
  {
    foreach (QSslError error, errors)
      std::cerr << "SSL Error: " << qPrintable(error.errorString()) << std::endl;

    wroteSslWarning = true;  // always write at least the first SSL errors
  }
}

//--------------------------------------------------------------------------------

bool Session::dataToWbXml(const QByteArray &data, QByteArray &result)
{
  WBXMLTree *tree = 0;

  WBXMLError error =
    wbxml_tree_from_xml(const_cast<WB_UTINY*>(reinterpret_cast<const WB_UTINY*>(data.constData())),
                        data.length(), &tree);

  if ( error != WBXML_OK )
  {
    qWarning("-error- wbxml_tree_from_xml:" + QByteArray::number(error) +
             QByteArray(reinterpret_cast<const char*>(wbxml_errors_string(error))));
    return false;
  }

  WB_UTINY *wbxml = 0;
  WB_ULONG wbxml_len = 0;
  WBXMLGenWBXMLParams params;
  params.wbxml_version = WBXML_VERSION_13;
  params.keep_ignorable_ws = 0;
  params.use_strtbl = 1;
  params.produce_anonymous = 1;  // important!
  error = wbxml_tree_to_wbxml(tree, &wbxml, &wbxml_len, &params);

  wbxml_tree_destroy(tree);

  if ( error != WBXML_OK )
  {
    qWarning("-error- wbxml_tree_to_wbxml:" + QByteArray::number(error) +
             QByteArray(reinterpret_cast<const char*>(wbxml_errors_string(error))));
    return false;
  }

  result = QByteArray(reinterpret_cast<const char*>(wbxml), wbxml_len);
  free(wbxml);

  return true;
}

//--------------------------------------------------------------------------------

bool Session::dataFromWbXml(const QByteArray &data, QByteArray &result)
{
  WBXMLGenXMLParams params;
  params.gen_type = WBXML_GEN_XML_INDENT;
  params.lang = WBXML_LANG_AIRSYNC;
  params.indent = 2;
  params.keep_ignorable_ws = 0;

  WB_UTINY *xml = 0;
  WB_ULONG len = 0;

  WBXMLError error =
    wbxml_conv_wbxml2xml_withlen(const_cast<WB_UTINY*>(reinterpret_cast<const WB_UTINY*>(data.constData())),
                                 data.length(), &xml, &len, &params);

  if ( error != WBXML_OK )
  {
    qWarning("-error- wbxml_conv_wbxml2xml_withlen:" + QByteArray::number(error) +
             QByteArray(reinterpret_cast<const char*>(wbxml_errors_string(error))));
    qWarning("input was: %s", data.toPercentEncoding().constData());
    return false;
  }

  result = QByteArray(reinterpret_cast<const char*>(xml), len);
  free(xml);

  return true;
}

//--------------------------------------------------------------------------------
