/* Copyright 2013, Martin Koller, kollix@aon.at

  This file is part of airsyncDownload.

  airsyncDownload is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  airsyncDownload is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with airsyncDownload.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QCoreApplication>
#include <QFile>
#include <QFileInfo>
#include <QDomDocument>
#include <QDateTime>
#include <QSettings>
#include <QProcess>
#include <QUuid>

#include <iostream>

#include <Session.hxx>

//--------------------------------------------------------------------------------

bool debug = false;
bool verbose = false;

//--------------------------------------------------------------------------------

class Lockfile
{
  public:
    Lockfile(const QString &mailbox) : lockfile(mailbox + ".lock") { }

    ~Lockfile() { unlock(); }

    bool lock()
    {
      QStringList args;
      args << QLatin1String("-r5")
           << QString::fromLocal8Bit(QFile::encodeName(lockfile));

      int rc = QProcess::execute(QLatin1String("lockfile"), args );
      return (rc == 0);
    }

    void unlock()
    {
      QFile(lockfile).remove();
    }

  private:
    QString lockfile;
};

//--------------------------------------------------------------------------------
// download new mails to a temporary mailbox file, then try to lock the final
// mailbox file and if succeeded, append all new mails to the final mbox
// (The lock shall be kept as short as possible, else e.g. kmail would force remove the lock ...)
// If locking fails, keep the tmp file so that the next invokation of airsyncDownload
// will still be able to retry again

int downloadFolder(Session &session, const QByteArray &collectionId, bool deleteMails, const QString &mboxfile)
{
  const QString tmpMailbox = mboxfile + ".tmp";
  QFile mailbox(tmpMailbox);
  if ( !mailbox.open(QIODevice::ReadWrite | QIODevice::Append) )
  {
    std::cerr << "ERROR opening tmp mbox file " << qPrintable(tmpMailbox) << std::endl
              << qPrintable(mailbox.errorString()) << std::endl;
    return -1;
  }

  QByteArray data;
  QDomDocument doc;

  // getting synckey for inbox mails
  data = "<Sync>"
           "<Collections>"
             "<Collection>"
               "<SyncKey>0</SyncKey>"
               "<CollectionId>" + collectionId + "</CollectionId>"
             "</Collection>"
           "</Collections>"
         "</Sync>";
  data = session.sendCommand("Sync", data);
  doc.setContent(data);

  QByteArray collectionSyncKey =
     doc.firstChildElement("Sync")
        .firstChildElement("Collections")
        .firstChildElement("Collection")
        .firstChildElement("SyncKey").text().toUtf8();

  if ( debug )
    qWarning("collectionSyncKey=%s", qPrintable(collectionSyncKey));

  // getting mails
  data = "<Sync>"
           "<Collections>"
             "<Collection>"
               "<SyncKey>" + collectionSyncKey + "</SyncKey>"
               "<CollectionId>" + collectionId + "</CollectionId>"
               "<GetChanges/>"
             "</Collection>"
           "</Collections>"
         "</Sync>";
  data = session.sendCommand("Sync", data);
  doc.setContent(data);

  collectionSyncKey =
     doc.firstChildElement("Sync")
        .firstChildElement("Collections")
        .firstChildElement("Collection")
        .firstChildElement("SyncKey").text().toUtf8();

  if ( debug )
    qWarning("collectionSyncKey=%s", qPrintable(collectionSyncKey));

  // fetch each mail
  QDomNodeList list = doc.elementsByTagName("Add");

  if ( verbose )
    std::cerr << list.count() << " new mails to download" << std::endl;

  for (int i = 0; i < list.count(); i++)
  {
    QDomNode node = list.at(i);
    QByteArray mailId = node.firstChildElement("ServerId").text().toUtf8();

    data = "<ItemOperations"
                " xmlns:airsync='AirSync'"
                " xmlns:airsyncbase='http://synce.org/formats/airsync_wm5/airsyncbase'"
                " xmlns='ItemOperations'>"
             "<Fetch xmlns='http://synce.org/formats/airsync_wm5/itemoperations'>"
               "<Store>Mailbox</Store>"
               "<airsync:CollectionId>" + collectionId + "</airsync:CollectionId>"
               "<airsync:ServerId>" + mailId + "</airsync:ServerId>"
               "<Options>"
                 "<airsync:MIMESupport>2</airsync:MIMESupport>"
                 "<airsyncbase:BodyPreference>"
                   "<airsyncbase:Type>4</airsyncbase:Type>"
                 "</airsyncbase:BodyPreference>"
               "</Options>"
             "</Fetch>"
           "</ItemOperations>";

    data = session.sendCommand("ItemOperations", data);

    QDomDocument mailDoc;
    if ( mailDoc.setContent(data) )
    {
      QString fullMail =
        mailDoc.firstChildElement("ItemOperations")
               .firstChildElement("Response")
               .firstChildElement("Fetch")
               .firstChildElement("Properties")
               .firstChildElement("Body")
               .firstChildElement("Data").text();

      if ( fullMail.isEmpty() )
      {
        std::cerr << "mail is empty (mailId=" << mailId.constData() << ") There's probably something going wrong!!" << std::endl;
        continue;
      }

      mailbox.write(QByteArray("From airsyncDownload@local.local ") +
                    QDateTime::currentDateTime().toString(Qt::ISODate).toUtf8() + "\n");
      mailbox.write(fullMail.toUtf8());
      mailbox.write("\n\n");
      mailbox.flush();

      if ( verbose )
        std::cerr << "mail #" << i << " added to mbox" << std::endl;

      if ( deleteMails )
      {
        data = "<Sync>"
                 "<Collections>"
                   "<Collection>"
                     "<SyncKey>" + collectionSyncKey + "</SyncKey>"
                     "<CollectionId>" + collectionId + "</CollectionId>"
                     "<DeletesAsMoves>0</DeletesAsMoves>"
                     "<Commands>"
                       "<Delete>"
                         "<ServerId>" + mailId + "</ServerId>"
                       "</Delete>"
                     "</Commands>"
                   "</Collection>"
                 "</Collections>"
               "</Sync>";

        data = session.sendCommand("Sync", data);

        QDomDocument doc;
        doc.setContent(data);

        collectionSyncKey =
           doc.firstChildElement("Sync")
              .firstChildElement("Collections")
              .firstChildElement("Collection")
              .firstChildElement("SyncKey").text().toUtf8();

        if ( debug )
          qWarning("collectionSyncKey=%s", qPrintable(collectionSyncKey));

        if ( verbose )
          std::cerr << "mail #" << i << " deleted from server" << std::endl;
      }
    }
  }

  mailbox.flush();

  // now append all downloaded mails (complete tmpMailbox) to the real mailbox file
  // to keep the file lock as short as possible
  Lockfile lockfile(mboxfile);
  if ( !lockfile.lock() )
  {
    // no lock possible, still keep the tmp file so we can try the next time
    std::cerr << "ERROR could not lock mbox file " << qPrintable(mboxfile) << std::endl
              << "I'll keep the downloaded mails in " << qPrintable(tmpMailbox) << std::endl
              << "Please restart airsyncDownload when the lock is possible" << std::endl
              << "and this mails will be appended to the mbox file" << std::endl;
    return -2;
  }

  QFile finalMailbox(mboxfile);
  if ( !finalMailbox.open(QIODevice::Append) )
  {
    std::cerr << "ERROR opening mbox file " << qPrintable(mboxfile) << std::endl
              << qPrintable(finalMailbox.errorString()) << std::endl;
    return -1;
  }

  mailbox.seek(0);

  while ( !(data = mailbox.read(1024 * 512)).isEmpty() )
    finalMailbox.write(data);

  finalMailbox.close();
  mailbox.close();
  mailbox.remove();  // we moved all new mails to final mbox - delete tmp file

  return 0;
}

//--------------------------------------------------------------------------------

int main(int argc, char **argv)
{
  QCoreApplication app(argc, argv);

  QSettings settings(QSettings::IniFormat, QSettings::UserScope, "airsyncDownload", "config");

  // create an initial file so that the user knows what options are available
  if ( !QFileInfo(settings.fileName()).exists() )
  {
    settings.setValue("serverUrl", "");
    settings.setValue("domain", "");
    settings.setValue("username", "");
    settings.setValue("password", "");
    settings.setValue("mboxfile", "allmails.mbox");
    settings.setValue("keepMailsOnServer", false);
  }

  QString serverUrl = settings.value("serverUrl", "").toString();
  QString domain    = settings.value("domain", "").toString();
  QString username  = settings.value("username", "").toString();
  QString password  = settings.value("password", "").toString();
  QString mboxfile  = settings.value("mboxfile", "allmails.mbox").toString();
  bool keepMailsOnServer = settings.value("keepMailsOnServer", false).toBool();

  if ( !settings.contains("deviceId") )
    settings.setValue("deviceId", QUuid::createUuid().toString().replace('-', "").mid(1, 32));

  QString uniqueDeviceId = settings.value("deviceId").toString();

  bool options = false;
  bool showUsage = false;
  bool allFolders = false;
  QByteArray folderName = "Inbox";

  for (int i = 1; i < argc; i++)
  {
    if ( strcmp(argv[i], "--debug") == 0 )
      debug = true;
    else if ( strcmp(argv[i], "--verbose") == 0 )
      verbose = true;
    else if ( strcmp(argv[i], "--options") == 0 )
      options = true;
    else if ( strcmp(argv[i], "--allFolders") == 0 )
      allFolders = true;
    else if ( strcmp(argv[i], "--folder") == 0 )
    {
      i++;
      if ( i < argc )
      {
        folderName = argv[i];
        if ( folderName != "Inbox" )
          mboxfile = QString::fromUtf8(folderName);
      }
      else
      {
        std::cerr << "folder name missing" << std::endl;
        return -1;
      }
    }
    else
      showUsage = true;
  }

  if ( showUsage )
  {
    std::cerr << "Usage: airsyncDownload [options]" << std::endl
              << "  This is a program to download mails from an Exchange Server," << std::endl
              << "  store them locally in an mbox file and optionally delete" << std::endl
              << "  the mails on the server."                                    << std::endl
              << std::endl
              << "  --help          ... this message" << std::endl
              << "  --options       ... send an OPTIONS request to the server and" << std::endl
              << "                      show its return values" << std::endl
              << "  --verbose       ... show progress information" << std::endl
              << "  --debug         ... turn on debugmode; shows the communication traffic" << std::endl
              << "  --folder <name> ... download only given mail folder; default: Inbox" << std::endl
              << "  --allFolders    ... download all mail folders into separate mbox files" << std::endl
              << "                      each named as the folder to download" << std::endl
              << std::endl
              << "Settings like serverUrl, username, etc. are stored in the config file" << std::endl
              << qPrintable(settings.fileName()) << std::endl << std::endl;
    return -1;
  }

  if ( serverUrl.isEmpty() )
  {
    std::cerr << "serverUrl is not specified. Check your config file" << std::endl
              << qPrintable(settings.fileName()) << std::endl;
    return -1;
  }

  if ( domain.isEmpty() )
  {
    std::cerr << "domain is not specified. Check your config file" << std::endl
              << qPrintable(settings.fileName()) << std::endl;
    return -1;
  }

  if ( username.isEmpty() )
  {
    std::cerr << "username is not specified. Check your config file" << std::endl
              << qPrintable(settings.fileName()) << std::endl;
    return -1;
  }

  if ( password.isEmpty() )
  {
    std::cerr << "password is not specified. Check your config file" << std::endl
              << qPrintable(settings.fileName()) << std::endl;
    return -1;
  }


  Session session(serverUrl, domain, username, password, uniqueDeviceId);
  session.setDebug(debug);

  if ( options )
  {
    session.options();
    return 0;
  }

  // get PolicyKey
  {
    QByteArray data =
          "<Provision xmlns=\"Provision:\">"
            "<Policies>"
              "<Policy>"
                "<PolicyType>MS-EAS-Provisioning-WBXML</PolicyType>"
              "</Policy>"
            "</Policies>"
          "</Provision>";

    data = session.sendCommand("Provision", data);

    QDomDocument doc;
    if ( !doc.setContent(data) )
    {
      qWarning("ERROR: could not parse XML data");
      return -1;
    }

    QDomNodeList list = doc.elementsByTagName("PolicyKey");
    if ( list.isEmpty() )
    {
      qWarning("ERROR: PolicyKey not found");
      return -1;
    }

    uint policyKey = list.at(0).firstChild().nodeValue().toUInt();

    session.setPolicyKey(policyKey);

    data = "<Provision xmlns=\"Provision:\">"
            "<Policies>"
              "<Policy>"
                "<PolicyType>MS-EAS-Provisioning-WBXML</PolicyType>\n"
                "<PolicyKey>" + QByteArray::number(policyKey) + "</PolicyKey>\n"
                "<Status xmlns=\"http://synce.org/formats/airsync_wm5/provision\">1</Status>\n"  // xmlns wichtig !!!!!
              "</Policy>"
            "</Policies>"
          "</Provision>";
    data = session.sendCommand("Provision", data);

    // set the final session policyKey
    if  ( !doc.setContent(data) )
    {
      std::cerr << "ERROR: could not parse XML data";
      return -1;
    }

    list = doc.elementsByTagName("PolicyKey");
    if ( list.isEmpty() )
    {
      std::cerr << "ERROR: PolicyKey not found" << std::endl;
      return -1;
    }

    session.setPolicyKey(list.at(0).firstChild().nodeValue().toUInt());
  }

  // get list of available folders
  QByteArray data = "<FolderSync xmlns=\"FolderHierarchy:\">\n"
    "<SyncKey xmlns=\"http://synce.org/formats/airsync_wm5/folderhierarchy\">0</SyncKey>\n"
    "</FolderSync>\n";
  data = session.sendCommand("FolderSync", data);

  QDomDocument doc;
  if ( !doc.setContent(data) )
  {
    std::cerr << "ERROR: could not parse XML data" << std::endl;
    return -1;
  }

  QDomNodeList list = doc.elementsByTagName("Add");

  QByteArray collectionId;
  for (int i = 0; i < list.count(); i++)
  {
    QDomNode node = list.at(i);

    if ( allFolders )
    {
      unsigned int type = node.firstChildElement("Type").text().toUInt();
      // MS-ASCMD: 2..6, 12 ... mail folders
      if ( ((type >= 2) && (type <= 6)) || (type == 12) )
      {
        QString fileName = node.firstChildElement("DisplayName").text();

        if ( verbose )
          std::cerr << "checking folder '" << qPrintable(fileName) << "'" << std::endl;

        int ret = downloadFolder(session, node.firstChildElement("ServerId").text().toUtf8(), !keepMailsOnServer, fileName);
        if ( ret != 0 )
          return ret;
      }
    }
    else  // only 1 folder
    {
      if ( node.firstChildElement("DisplayName").text().toUtf8() == folderName )
      {
        collectionId = node.firstChildElement("ServerId").text().toUtf8();
        break;
      }
    }
  }

  if ( allFolders )
    return 0;

  if ( collectionId.isEmpty() )
  {
    std::cerr << "Folder '" << folderName.constData() << "' not found";
    return -1;
  }

  if ( debug )
    qWarning("%s has collection ServerId=%s", folderName.constData(), qPrintable(collectionId));

  return downloadFolder(session, collectionId, !keepMailsOnServer, mboxfile);
}
